// ---------------------------------------------------------------------
/// \file airquality-2-click.c - A small example with the WaRP7 Board and the Air Quality 2 click.
//
/// \author Pierre-Jean Texier 2017 - <texier.pj2@gmail.com>
//
/// \licence MIT.
//

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <getopt.h>
#include "src/i2c.h"

// Device
#define IAQ_I2C_ADDR				(0x5A)
#define IAQCORE_DATA_SIZE			9
// Status codes
#define IAQCORE_STATUS_OK			(0x00)
#define IAQCORE_STATUS_RUNIN			(0x10)
#define IAQCORE_STATUS_BUSY			(0x01)
#define IAQCORE_STATUS_ERROR			(0x80)

// Rx Buffer
uint8_t data[IAQCORE_DATA_SIZE] = { 0 };

// i2c
i2c_t i2c;

static void display_help(const char * name)
{
    	fprintf(stderr, "usage: %s options...\n", name);
	fprintf(stderr, " options:\n");
	fprintf(stderr, " -c --co2\n");
	fprintf(stderr, " -s --status \n");
	fprintf(stderr, " -r --resistance\n");
	fprintf(stderr, " -t --tvoc\n");
	fprintf(stderr, " -h --help Prints this help\n\n");
	fprintf(stderr, "Example: %s --co2 --tvoc\n\n", name);
}

void readFromSensor()
{	
	struct i2c_msg msgs[2] =
        {        
            { .addr = IAQ_I2C_ADDR, .flags = I2C_M_RD, .len = 9, .buf = data},
        };
        
	if (i2c_transfer(&i2c, msgs, 2) < 0) {
		fprintf(stderr, "i2c_transfer(): %s\n", i2c_errmsg(&i2c));
		exit(1);
	}
} 

uint8_t getStatus(void) {
	return data[2];
}

uint16_t getCO2Prediction(void) {
    	return data[0] << 8 | data[1];
}

uint32_t getSensorResistance(void) {
    	return data[3] << 24 | data[4] << 16 | data[5] << 8 | data[6];
}

uint16_t getTVOCPrediction(void) {
	return data[7] << 8 | data[8];
}

static int parse_opts(int argc, char *argv[])
{	
	int long_index = 0;
	int opt;
	
	static struct option option[] = 
	{
		{"co2", 	no_argument, 		NULL, 'c' },		
		{"tvoc", 	no_argument, 		NULL, 't' },
		{"resistance", 	no_argument, 		NULL, 'r' },
		{"status", 	no_argument, 		NULL, 's' },
		{"help", 	no_argument, 		NULL, 'h' },
		{0, 			0, 		0, 	0 }
	};
		
 	
	while ((opt = getopt_long(argc, argv, "ctrsh", option, &long_index)) >= 0) {
		switch(opt) {
			case 'c':
				printf("%d\n", getCO2Prediction());
				break;
			case 't':
				printf("%d\n", getTVOCPrediction());
				break;
			case 'r':	
				printf("%d\n", getSensorResistance());
				break;
			case 's':
				switch(getStatus()) 
				{
					case IAQCORE_STATUS_OK:
						printf("OK\n");
						break;
					case IAQCORE_STATUS_RUNIN:
						printf("RUNIN\n");
					    	break;
					case IAQCORE_STATUS_BUSY:
					    	printf("BUSY\n");
					    	break;
					case IAQCORE_STATUS_ERROR:
					    	printf("ERROR!\n");
					    	break;
					default:
					    	printf("Undefined");
					    	break;
				}
				break;
			case 'h':
				display_help(argv[0]);
				exit(EXIT_SUCCESS);	
			default:
				fprintf(stderr,"[ERROR] %s: Bad option. -h for help\n", argv[0]);
				exit(EXIT_FAILURE);
		}				
	}
	
	return 0;
}

int main(int argc, char * argv[])
{
	int ret_val = 0;
	
    	if (i2c_open(&i2c, "/dev/i2c-2") < 0) {
        	fprintf(stderr, "i2c_open(): %s\n", i2c_errmsg(&i2c));
        	exit(1);
    	}	   		
	  	
  	if (argc < 2){
		display_help(argv[0]);
	}
	
	// Get All values from iaq
	readFromSensor();
		
	ret_val = parse_opts(argc, argv);
	if (ret_val) {
		display_help(argv[0]);
		return ret_val;
	}

	// Close i2c
    	i2c_close(&i2c);

    	return 0;
}
